//============================================================================
// Wo
//
// (c)2023 Maikon-Giken
//============================================================================
#include <stdio.h>
#include <string.h>

//============================================================================
// `.
//============================================================================

//============================================================================
// vg^Cv.
//============================================================================

//============================================================================
// f[^.
//============================================================================
FILE	*stdin, *stdout;
static	char stdio_name[STDIO_FILES_MAX][STDIO_FNAME_MAX];
static	STDIO_OPEN_FUNCTION stdio_fopen[STDIO_FILES_MAX];
static	int file_max;		// o^hCo

//============================================================================
// API֐.
//============================================================================
int		STDIO_set_driver(const char *name, STDIO_OPEN_FUNCTION func) {
	if (file_max < STDIO_FILES_MAX) {
		strcpy(stdio_name[file_max], name);
		stdio_fopen[file_max++] = func;
		return 0;
	}
	return EOF;
}

//---------------------------------------------------------------------
FILE *	STDIO_fopen(const char *name, const char *mode) {
	char *p;
	int i, n;
	
	p = strchr(name, ':');
	n = (int)p - (int)name + 1;
	if (p++ == NULL)  return NULL;
	
	for (i = 0; i < file_max; i++) {
		if (strncmp(name, stdio_name[i], n) == 0) {
			if (stdio_fopen[i] == NULL) return NULL;
			return stdio_fopen[i](p, mode);
		}
	}
	return NULL;
}

//---------------------------------------------------------------------
int		STDIO_fclose(FILE *fp) {
	if (fp == NULL || fp->close == NULL) return EOF;
	return fp->close(fp);
}

//---------------------------------------------------------------------
int		STDIO_fputc(int c, FILE *fp) {

	if (fp == NULL) return EOF;

#if   STDIO_VERSION == 1
	*fp->wp = c;
	if (fp->wp < fp->end) fp->wp++;
	else fp->wp = fp->begin;

#elif STDIO_VERSION == 2
	fp->buff[fp->wp++] = c;
#endif

	if (fp->wake != NULL) fp->wake();
	return c;
}

//---------------------------------------------------------------------
int		STDIO_fgetc(FILE *fp) {
	int c;
	
	if (fp == NULL || fp->rp == fp->wp) return EOF;

#if   STDIO_VERSION == 1
	c = *fp->rp;
	if (fp->rp < fp->end) fp->rp++;
	else fp->rp = fp->begin;

#elif STDIO_VERSION == 2
	c = fp->buff[fp->rp++];
#endif

	fp->cp = fp->rp;
	return c;
}

//---------------------------------------------------------------------
int		STDIO_puts(const char *s) {
	if (fputs(s, stdout) == EOF) return EOF;
	if (fputc('\n', stdout) == EOF) return EOF;
	return 0;
}

//---------------------------------------------------------------------
int		STDIO_fputs(const char *s, FILE *fp) {
	if (fp == NULL) return EOF;
	while (*s != '\0') {

#if   STDIO_VERSION == 1
		*fp->wp = *s++;
		if (fp->wp < fp->end) fp->wp++;
		else fp->wp = fp->begin;

#elif STDIO_VERSION == 2
		fp->buff[fp->wp++] = *s++;
#endif

	}
	if (fp->wake != NULL) fp->wake();
	return 0;
}

//---------------------------------------------------------------------
char*	STDIO_gets(char *s) {
	char *p;
	
#if   STDIO_VERSION == 1
	if (fgets(s, (int)(stdin->end - stdin->begin), stdin) == NULL) return NULL;

#elif STDIO_VERSION == 2
	if (fgets(s, 255, stdin) == NULL) return NULL;
#endif

	p = strchr(s, '\0');
	p--;
	if (*p == '\n') *p = '\0';
	return s;
}

//---------------------------------------------------------------------
char*	STDIO_fgets(char *s, int n, FILE *fp) {
	int c;
	char* p;

	if (fp == NULL) return NULL;

#if   STDIO_VERSION == 1
	p = s;
	while (fp->cp != fp->wp) {
		c = *fp->cp; if (fp->cp < fp->end) fp->cp++; else fp->cp = fp->begin;
		if (c == '\n') {
			while (n > 1) {
				c = *fp->rp; if (fp->rp < fp->end) fp->rp++; else fp->rp = fp->begin;
				if (fp->binary != 'b' && c == '\b') {
					if (p != s) p--, n++;
				}
				else {
					*p++ = c;
					n--;
					if (c == '\n') break;
				}
			}
			*p = '\0';
			fp->cp = fp->rp;
			return s;
		}
	}

#elif STDIO_VERSION == 2
	p = s;
	while (fp->cp != fp->wp) {
		c = fp->buff[fp->cp++];
		if (c == '\n') {
			while (n > 1) {
				c = fp->buff[fp->rp++];
				if (fp->binary != 'b' && c == '\b') {
					if (p != s) p--, n++;
				}
				else {
					*p++ = c;
					n--;
					if (c == '\n') break;
				}
			}
			*p = '\0';
			fp->cp = fp->rp;
			return s;
		}
	}
#endif

	return NULL;
}

#ifdef STDIO_USE_PRINTF
//---------------------------------------------------------------------
int		STDIO_printf(const char *s, ...) {
	char buff[STDIO_BUFFER_MAX];
	int n;
	va_list ap;

	va_start(ap, s);
	n = vsprintf(buff, s, ap);
	va_end(ap);

	if (fputs(buff, stdout) == EOF) return EOF;
	return n;
}

//---------------------------------------------------------------------
int		STDIO_fprintf(FILE *fp, const char *s, ...) {
	char buff[STDIO_BUFFER_MAX];
	int n;
	va_list ap;

	va_start(ap, s);
	n = vsprintf(buff, s, ap);
	va_end(ap);

	if (fputs(buff, fp) == EOF) return EOF;
	return n;
}

//---------------------------------------------------------------------
int		STDIO_sprintf(char *wk, const char *s, ...) {
	int n;
	va_list ap;

	va_start(ap, s);
	n = vsprintf(wk, s, ap);
	va_end(ap);

	return n;
}

#endif

#ifdef STDIO_USE_SCANF
//static char scan_buff[STDIO_BUFFER_MAX];
//---------------------------------------------------------------------
int		STDIO_scanf(const char *s, ...) {
	char scan_buff[STDIO_BUFFER_MAX];
	int n;
	va_list ap;

//	  if (scan_buff[0] == '\0') {
		if (fgets(scan_buff, STDIO_BUFFER_MAX, stdin) == NULL) return EOF;
//	  }
	va_start(ap, s);
	n = vsscanf(scan_buff, s, ap);
	va_end(ap);
//	  if (n == EOF) scan_buff[0] = '\0';
	return n;
}

//---------------------------------------------------------------------
int		STDIO_fscanf(FILE *fp, const char *s, ...) {
	char scan_buff[STDIO_BUFFER_MAX];
	int n;
	va_list ap;

//	  if (scan_buff[0] == '\0') {
		if (fgets(scan_buff, STDIO_BUFFER_MAX, fp) == NULL) return EOF;
//	  }
	va_start(ap, s);
	n = vsscanf(scan_buff, s, ap);
	va_end(ap);
//	  if (n == EOF) scan_buff[0] = '\0';
	return n;
}

//---------------------------------------------------------------------
int		STDIO_sscanf(char *wk, const char *s, ...) {
	int n;
	va_list ap;

	va_start(ap, s);
	n = vsscanf(wk, s, ap);
	va_end(ap);
	
	return n;
}

//---------------------------------------------------------------------
int		vsscanf(const char *wk, const char *s, va_list ap) {
	/* CC-RHŌT|[gĂȂ */
	return 0;
}
#endif

//============================================================================
// ^XNA又.
//============================================================================

//============================================================================
// [J֐.
//============================================================================

//============================================================================
// nh.
//============================================================================

//============================================================================
