//============================================================================
// RS485^XN
//
// (c)2023 Maikon-Giken
//============================================================================
#include "kernel.h"
#include "task_rs485.h"
#include <stdio.h>

//============================================================================
// `.
//============================================================================
#if RS485_BAUDRATE == 921600
	#define kRS485_LBRP01	(4 -1)
	#define kRS485_LWBR		((10 << 4) | (0 << 1))
#elif RS485_BAUDRATE == 460800
	#define kRS485_LBRP01	(8 -1)
	#define kRS485_LWBR		((10 << 4) | (0 << 1))
#elif RS485_BAUDRATE == 230400
	#define kRS485_LBRP01	(16 -1)
	#define kRS485_LWBR		((10 << 4) | (0 << 1))
#elif RS485_BAUDRATE == 115200
	#define kRS485_LBRP01	(32 -1)
	#define kRS485_LWBR		((10 << 4) | (0 << 1))
#endif

#define RS485_TX_REQ	(1)
#define RS485_TX_NON	(0)

//============================================================================
// vg^Cv.
//============================================================================
void RS485_main(void);
void RS485_wake(void);

void RS485_RX_Handler(void);
void RS485_TX_Handler(void);

//============================================================================
// f[^.
//============================================================================
static uchar RS485_tx_buff[RS485_SEND_SIZE];	// MOEobt@.
static uchar RS485_rx_buff[RS485_RECV_SIZE];	// MOEobt@.
static FILE RS485_txf = {						// MFILE
	RS485_tx_buff, RS485_tx_buff, RS485_tx_buff, RS485_tx_buff,
	RS485_tx_buff + (RS485_SEND_SIZE - 1), 0,0,0,0,
	RS485_wake, RS485_close
};
static FILE RS485_rxf = {						// MFILE
	RS485_rx_buff, RS485_rx_buff, RS485_rx_buff, RS485_rx_buff,
	RS485_rx_buff + (RS485_RECV_SIZE - 1), 0,0,0,0,
	NULL, RS485_close
};

static int RS485_over_error;
static int RS485_rec_error;
static schar RS485_tx_req;

//============================================================================
// API֐.
//============================================================================
// 
void RS485_init(void) {
	//RS485ݒ
	RS485_LWBR = kRS485_LWBR;		//Tv/bit, vXP[
	RS485_LBRP01 = kRS485_LBRP01;	//{[[g
	RS485_LMD = ((0 << 5) | 1);		//MtB^ARS485[h
	RS485_LEDE = (0x0c);			//G[o
	RS485_LBFC = (0);				//tH[}bg
	RS485_LCUC = (1);				//RS485N	
	while(!RS485_LMST);
	
	//|[gݒ
	RS485_TX_PIN_PFC = 1;
	RS485_TX_PIN_PMC = 1;
	RS485_TX_PIN_PM = 0;
	
	RS485_RX_PIN_PU = 1;
	RS485_RX_PIN_PFCE = 1;
	RS485_RX_PIN_PFCAE = 1;
	RS485_RX_PIN_PMC = 1;

	RS485_TE_PIN = 0;
	RS485_TE_PIN_PM = 0;
	
	//MI̊ݓo^AvCIeBݒ肨ъ݋
	NVIC_set_vector(RS485_RX_IRQ, RS485_RX_ICH, RS485_RX_Handler);
	NVIC_set_priority(RS485_RX_IRQ, RS485_RX_IPR);		//Dxݒ.
	NVIC_clr_irq(RS485_RX_IRQ);							//NVICݗvNA.
//	NVIC_disable_irq(RS485_RX_IRQ);						//NVICݕs.

	//MI̊ݓo^AvCIeBݒ肨ъ݋
	NVIC_set_vector(RS485_TX_IRQ, RS485_TX_ICH, RS485_TX_Handler);
	NVIC_set_priority(RS485_TX_IRQ, RS485_TX_IPR);		//Dxݒ.
	NVIC_clr_irq(RS485_TX_IRQ);							//NVICݗvNA.
//	NVIC_disable_irq(RS485_TX_IRQ);						//NVICݕs.

	RS485_LUOER = ((1 << 1) | 1);	//M
	start_task(RS485_taskid, RS485_main);
}
//---------------------------------------------------------------------
// RS485I[v֐
FILE* RS485_open(const char *name, const char *mode) {
	if (*mode == 'w') {
		if (RS485_txf.rw_mode)	 return NULL;
		RS485_txf.wp = RS485_tx_buff;
		RS485_txf.rp = RS485_tx_buff;
		RS485_txf.cp = RS485_tx_buff;
		RS485_txf.rw_mode = mode[0];
		RS485_txf.binary = mode[1];
		return &RS485_txf;
	}
	else if (*mode == 'r') {
		if (RS485_rxf.rw_mode)	 return NULL;
		RS485_rxf.wp = RS485_rx_buff;
		RS485_rxf.rp = RS485_rx_buff;
		RS485_rxf.cp = RS485_rx_buff;
		RS485_rxf.rw_mode = mode[0];
		RS485_rxf.binary = mode[1];
		NVIC_enable_irq(RS485_RX_IRQ);
		return &RS485_rxf;
	}
	return NULL;
}

//---------------------------------------------------------------------
// RS485N[Y֐
int RS485_close(void *fp) {
	if (fp == &RS485_txf) {
		RS485_txf.rw_mode = '\0';
		NVIC_disable_irq(RS485_TX_IRQ);
		RS485_tx_req = RS485_TX_NON;
		return 0;
	}
	else if (fp == &RS485_rxf) {
		RS485_rxf.rw_mode = '\0';
		NVIC_disable_irq(RS485_RX_IRQ);
		return 0;
	}
	return EOF;
}

//---------------------------------------------------------------------
// MJn (ԐAPI)
void RS485_wake(void) {
	__DI();
	if (RS485_txf.rw_mode) {
		if (RS485_tx_req == RS485_TX_NON) {
			RS485_tx_req = RS485_TX_REQ;
			wakeup_task(RS485_taskid);
		}
	}
	__EI();
	return;
}

//============================================================================
// ^XNA又.
//============================================================================
void RS485_main(void) {
	__DI();
	if (RS485_tx_req) {
		if(RS485_TE_PIN == 0) {
			if (RS485_LST_RECV) { set_retry_task(); }
			else {
				RS485_TE_PIN = 1;
				NVIC_set_irq(RS485_TX_IRQ);
				NVIC_enable_irq(RS485_TX_IRQ);
			}
		}
	}
	else {
		if(RS485_TE_PIN == 1) {
			if (RS485_LST_SEND) { set_retry_task(); }
			else {
				RS485_TE_PIN = 0;
				NVIC_disable_irq(RS485_TX_IRQ);
			}
		}
	}
	__EI();
}

//============================================================================
// [J֐.
//============================================================================

//============================================================================
// nh.
//============================================================================
#pragma interrupt RS485_RX_Handler(enable=false, channel=122, fpu=false, callt=false)
void RS485_RX_Handler(void) {

	*RS485_rxf.wp = RS485_RXD;
	if (RS485_ERROR) {
		RS485_rec_error++;
		RS485_ERROR = 0;
	}
	if (RS485_rxf.wp < RS485_rxf.end) RS485_rxf.wp++; else RS485_rxf.wp = RS485_rxf.begin;
	if (RS485_rxf.wp == RS485_rxf.rp) RS485_over_error++;
}

#pragma interrupt RS485_TX_Handler(enable=true, channel=121, fpu=false, callt=false)
void RS485_TX_Handler(void) {

	if (RS485_txf.rp != RS485_txf.wp) {
		RS485_TXD = *RS485_txf.rp;
		if (RS485_txf.rp < RS485_txf.end) RS485_txf.rp++; else RS485_txf.rp = RS485_txf.begin;
	}
	else {
		NVIC_disable_irq(RS485_TX_IRQ);
		RS485_tx_req = RS485_TX_NON;
		wakeup_task(RS485_taskid);
	}
}

//============================================================================
