//============================================================================
// UART^XN
//
// (c)2023 Maikon-Giken
//============================================================================
#include "kernel.h"
#include "task_uart.h"
#include <stdio.h>

//============================================================================
// `.
//============================================================================
#if UART_BAUDRATE == 921600
	#define kUART_LBRP01	(4 -1)
	#define kUART_LWBR		((10 << 4) | (0 << 1))
#elif UART_BAUDRATE == 460800
	#define kUART_LBRP01	(8 -1)
	#define kUART_LWBR		((10 << 4) | (0 << 1))
#elif UART_BAUDRATE == 230400
	#define kUART_LBRP01	(16 -1)
	#define kUART_LWBR		((10 << 4) | (0 << 1))
#elif UART_BAUDRATE == 115200
	#define kUART_LBRP01	(32 -1)
	#define kUART_LWBR		((10 << 4) | (0 << 1))
#endif

//============================================================================
// vg^Cv.
//============================================================================
void UART_wake(void);

void UART_RX_Handler(void);
void UART_TX_Handler(void);
void UART_ER_Handler(void);

//============================================================================
// f[^.
//============================================================================
static uchar UART_tx_buff[UART_SEND_SIZE];	// MOEobt@.
static uchar UART_rx_buff[UART_RECV_SIZE];	// MOEobt@.

static FILE UART_txf = {					// MFILE
	UART_tx_buff, UART_tx_buff, UART_tx_buff, UART_tx_buff,
	UART_tx_buff + (UART_SEND_SIZE - 1), 0,0,0,0,
	UART_wake, UART_close
};
static FILE UART_rxf = {					// MFILE
	UART_rx_buff, UART_rx_buff, UART_rx_buff, UART_rx_buff,
	UART_rx_buff + (UART_RECV_SIZE - 1), 0,0,0,0,
	NULL, UART_close
};

static int UART_over_error;
static int UART_rec_error;

//============================================================================
// API֐.
//============================================================================
// 
void UART_init(void) {
	//UARTݒ
	UART_LWBR = kUART_LWBR;			//Tv/bit, vXP[
	UART_LBRP01 = kUART_LBRP01;		//{[[g
	UART_LMD = ((0 << 5) | 1);		//MtB^AUART[h
	UART_LEDE = (0x0c);				//G[o
	UART_LBFC = (0);				//tH[}bg
	UART_LCUC = (1);				//UARTN	
	while(!UART_LMST);
	
	//|[gݒ
	UART_TX_PIN_PFC = 1;
	UART_TX_PIN_PMC = 1;
	UART_TX_PIN_PM = 0;
	
	UART_RX_PIN_PU = 1;
	UART_RX_PIN_PFCE = 1;
	UART_RX_PIN_PFCAE = 1;
	UART_RX_PIN_PMC = 1;
	
	//MI̊ݓo^AvCIeBݒ肨ъ݋
	NVIC_set_vector(UART_RX_IRQ, UART_RX_ICH, UART_RX_Handler);
	NVIC_set_priority(UART_RX_IRQ, UART_RX_IPR);		//Dxݒ.
	NVIC_clr_irq(UART_RX_IRQ);							//NVICݗvNA.
//	NVIC_disable_irq(UART_RX_IRQ);						//NVICݕs.

	//MI̊ݓo^AvCIeBݒ肨ъ݋
	NVIC_set_vector(UART_TX_IRQ, UART_TX_ICH, UART_TX_Handler);
	NVIC_set_priority(UART_TX_IRQ, UART_TX_IPR);		//Dxݒ.
	NVIC_clr_irq(UART_TX_IRQ);							//NVICݗvNA.
//	NVIC_disable_irq(UART_TX_IRQ);						//NVICݕs.

	//G[̊ݓo^AvCIeBݒ肨ъ݋
	NVIC_set_vector(UART_ER_IRQ, UART_ER_ICH, UART_ER_Handler);
	NVIC_set_priority(UART_ER_IRQ, UART_ER_IPR);		//Dxݒ.
//	NVIC_clr_irq(UART_ER_IRQ);							//NVICݗvNA.
	NVIC_enable_irq(UART_ER_IRQ);						//NVIC݋.

	UART_LUOER = ((1 << 1) | 1);	//M
}
//---------------------------------------------------------------------
// UARTI[v֐
FILE* UART_open(const char *name, const char *mode) {
	if (*mode == 'w') {
		if (UART_txf.rw_mode)	return NULL;
		UART_txf.wp = UART_tx_buff;
		UART_txf.rp = UART_tx_buff;
		UART_txf.cp = UART_tx_buff;
		UART_txf.rw_mode = mode[0];
		UART_txf.binary = mode[1];
		return &UART_txf;
	}
	else if (*mode == 'r') {
		if (UART_rxf.rw_mode)	return NULL;
		UART_rxf.wp = UART_rx_buff;
		UART_rxf.rp = UART_rx_buff;
		UART_rxf.cp = UART_rx_buff;
		UART_rxf.rw_mode = mode[0];
		UART_rxf.binary = mode[1];
		NVIC_enable_irq(UART_RX_IRQ);
		return &UART_rxf;
	}
	return NULL;
}

//---------------------------------------------------------------------
// UARTN[Y֐
int UART_close(void *fp) {
	if (fp == &UART_txf) {
		UART_txf.rw_mode = '\0';
		NVIC_disable_irq(UART_TX_IRQ);
		return 0;
	}
	else if (fp == &UART_rxf) {
		UART_rxf.rw_mode = '\0';
		NVIC_disable_irq(UART_RX_IRQ);
		return 0;
	}
	return EOF;
}

//---------------------------------------------------------------------
// MJn (ԐAPI)
void UART_wake(void) {
	__DI();
	if (UART_txf.rw_mode && NVIC_enable_check(UART_TX_IRQ)) {
		NVIC_set_irq(UART_TX_IRQ);
		NVIC_enable_irq(UART_TX_IRQ);
	}
	__EI();
	return;
}

//============================================================================
// ^XNA又.
//============================================================================

//============================================================================
// [J֐.
//============================================================================

//============================================================================
// nh.
//============================================================================
#pragma interrupt UART_RX_Handler(enable=false, channel=35, fpu=false, callt=false)
void UART_RX_Handler(void) {

	*UART_rxf.wp = UART_RXD;
	if (UART_ERROR) {
		UART_rec_error++;
		UART_ERROR = 0;
	}
	if (UART_rxf.wp < UART_rxf.end) UART_rxf.wp++; else UART_rxf.wp = UART_rxf.begin;
	if (UART_rxf.wp == UART_rxf.rp) UART_over_error++;
}

#pragma interrupt UART_TX_Handler(enable=true, channel=34, fpu=false, callt=false)
void UART_TX_Handler(void) {

	if (UART_txf.rp != UART_txf.wp) {
		UART_TXD = *UART_txf.rp;
		if (UART_txf.rp < UART_txf.end) UART_txf.rp++; else UART_txf.rp = UART_txf.begin;
	}
	else {
		NVIC_disable_irq(UART_TX_IRQ);
	}
}

#pragma interrupt UART_ER_Handler(enable=false, channel=36, fpu=false, callt=false)
void UART_ER_Handler(void) {
	UART_rec_error++;
}

//============================================================================
